#!/usr/bin/perl
# Convert HVM domains to PV on XenServer 4.1.0.  Have not tested on other Xenserver versions.
# Written by Randy McAnally.

use strict;
use Getopt::Long;

## Default values for XenServer 4.10 and Debian 5.03 DomU
my $cmd_xe = "/usr/bin/xe";
my $cmd_xenstore_ls = "/usr/bin/xenstore-ls";
my $root = 1;  ## Root partition on DomU
my $kernel;
my $ramdisk;

## Do not edit anything below this

my %globals;
my ($cmd,$name,$uuid,$vm,$disks);

usage() if ( !GetOptions( 'cmd=s' => \$cmd, 
                          'vm=s' => \$name, 
                          'uuid=s' => \$uuid, 
                          'root=i' => \$root, 
                          'kernel=s' => \$kernel, 
                          'ramdisk=s' => \$ramdisk, 
                          '<>' => \&usage) );
usage() if ( !$cmd || !($uuid || $name) );
usage() if ( !($cmd =~ m/^(hvmtopv|pvtohvm)$/));

my $ns = namespace() || die "Could not gather XenServer namespace";

if ( $name )
{
  if ( exists($ns->{'name'}->{$name}) )
  {
    print "uuid = ".$ns->{'name'}->{$name}->{'uuid'}."\n";
    print "cmd = ".$cmd."\n";
    $uuid = $ns->{'name'}->{$name}->{'uuid'};
    $vm = $ns->{'name'}->{$name};
  } else {
    print "VM not found\n";
  }
} else {
  if ( exists($ns->{'uuid'}->{$uuid}) )
  {
    print "name = ".$ns->{'uuid'}->{$uuid}->{'name-label'}."\n";
    print "cmd = ".$cmd."\n";
    $vm = $ns->{'uuid'}->{$uuid};
  } else {
    print "VM not found\n";
  }
}


my $params = get_vm_params($uuid) || die "Could not get params!";

if ( $cmd eq 'hvmtopv' )
{

    my $result;
    print "Changing HVM-boot-policy\n";
    if ( $result = xe_cmd('vm-param-set uuid='.$uuid.' HVM-boot-policy=""') ) { print "Error: $result\n"; exit; }
    my $disk; if ( !($disk = vm_disk_list($uuid)->{'Disk 0'}) ) { print "Error: Couldn't obtain VBD UUID\n"; exit; }
    my $pos; if ( $disk->{'userdevice'} == 0 ) { $pos = 'a'; } elsif ( $disk->{'userdevice'} == 1 ) { $pos = 'b'; } elsif ( $disk->{'userdevice'} == 2 ) { $pos = 'c'; } elsif ( $disk->{'userdevice'} == 3 ) { $pos = 'd'; } else { print "Error: Couldn't determine VBD device position\n"; exit; }
    print "Changing PV-args\n";
    if ( $result = xe_cmd('vm-param-set uuid='.$uuid.' PV-args="console=hvc0 xencons=hvc root=/dev/xvd'.$pos.$root.'"') ) { print "Error: $result\n"; exit; }
    print "Setting disk boot flag\n";
    if ( $result = xe_cmd('vbd-param-set uuid='.$disk->{'uuid'}.' bootable=true') ) { print "Error: $result\n"; exit; }
    #print "Changing PV-bootloader-args\n";
    #if ( $result = xe_cmd('vm-param-set uuid='.$uuid.' PV-bootloader-args="--kernel '.$kernel.' --ramdisk '.$ramdisk.'"') ) { print "Error: $result\n"; exit; }
    print "***Please update /etc/fstab, /etc/inittab. then reboot VM\n";
    
} elsif ( $cmd eq 'pvtohvm' ) {

    my $result;
    print "Changing HVM-boot-policy\n";
    if ( $result = xe_cmd('vm-param-set uuid='.$uuid.' HVM-boot-policy="BIOS order"') ) { print "Error: $result\n"; exit; }
    my $disk; if ( !($disk = vm_disk_list($uuid)->{'Disk 0'}) ) { print "Error: Couldn't obtain VBD UUID\n"; exit; }
    my $pos; if ( $disk->{'userdevice'} == 0 ) { $pos = 'a'; } elsif ( $disk->{'userdevice'} == 1 ) { $pos = 'b'; } elsif ( $disk->{'userdevice'} == 2 ) { $pos = 'c'; } elsif ( $disk->{'userdevice'} == 3 ) { $pos = 'd'; } else { print "Error: Couldn't determine VBD device position\n"; exit; }
    print "Changing PV-args\n";
    if ( $result = xe_cmd('vm-param-set uuid='.$uuid.' PV-args=""') ) { print "Error: $result\n"; exit; }
    print "Setting disk boot flag\n";
    if ( $result = xe_cmd('vbd-param-set uuid='.$disk->{'uuid'}.' bootable=false') ) { print "Error: $result\n"; exit; }
    print "Changing PV-bootloader-args\n";
    if ( $result = xe_cmd('vm-param-set uuid='.$uuid.' PV-bootloader-args=""') ) { print "Error: $result\n"; exit; }
    print "***Please update /etc/fstab, /etc/inittab. then reboot VM\n";

}
print "Done.\n";
exit;


#### Subroutines

sub usage
{
  print "Usage: vmtool.pl --cmd (hvmtopv|pvtohvm) ([--vm name-label] or [--uuid uuid]) [--root partition_num] [--kernel kernel_path] [--ramdisk initrd_path]\n";
  exit;
}

sub get_vm_params($)
{
  my $uuid = shift || die "Missing UUID";
  my $command = $cmd_xe." vm-param-list uuid=".$uuid." 2>&1|";
  my ($fh,$hash);
  open($fh,$command) || die "Could not execute:\n".$command." ".$!;
  
  ### Create uuid key hash
  while ( <$fh> )
  {
    chomp;
    if ( my ($name,$value) = split(/\(.*\)\s*?:/,$_) )
    {
      $name =~ s/(^\s*|\s*$)//g;
      $value =~ s/(^\s*|\s*$)//g;
      $hash->{$name} = $value;
    }
  }
  close $fh;
  return $hash;
}
  
  
sub namespace($)
{
  my $force = shift;
  return $globals{NAMESPACE} if ( !$force && $globals{NAMESPACE} );
  my $hash;
   
  $hash->{'host'} = host_namespace();
  
  ### Get VM info
  my ($fh,$command);
  my $command = $cmd_xe." vm-list 2>&1|";
  open($fh,$command) || die "Could not execute:\n".$command." ".$!;

  my $current_uuid;
  my $dom0_uuid;
  
  ### Create uuid key hash
  while ( <$fh> )
  {
    chomp;
    if ( my ($name,$value) = split(/\(.*\)\s*?:/,$_) )
    {
      $name =~ s/(^\s*|\s*$)//g;
      $value =~ s/(^\s*|\s*$)//g;
      if ( $name eq 'uuid' )
      {
        $current_uuid = $value;
        $hash->{'uuid'}->{$current_uuid}->{'uuid'} = $current_uuid;
      } else {
        $hash->{'uuid'}->{$current_uuid}->{$name} = $value;
      }
      ### Get domain-id if VM is running
      if ( $name eq 'power-state' && $value eq 'running' )
      {
        $hash->{'uuid'}->{$current_uuid}->{'dom-id'} = get_domain_id_from_uuid($current_uuid);
        if ( $hash->{'uuid'}->{$current_uuid}->{'dom-id'} == 0 )
        {
          $dom0_uuid = $current_uuid;
        }
      }
    }
  }
  close $fh;
  
  ### Create vm-name key hash
  foreach my $uuid ( keys %{$hash->{'uuid'}} )
  {
    foreach my $key ( keys %{$hash->{'uuid'}->{$uuid}} )
    {
      $hash->{'name'}->{$hash->{'uuid'}->{$uuid}->{'name-label'}}->{$key} = $hash->{'uuid'}->{$uuid}->{$key};
    }
  }
  
  ### Create dom-id key hash
  foreach my $uuid ( keys %{$hash->{'uuid'}} )
  {
    foreach my $key ( keys %{$hash->{'uuid'}->{$uuid}} )
    {
      $hash->{'dom-id'}->{$hash->{'uuid'}->{$uuid}->{'dom-id'}}->{$key} = $hash->{'uuid'}->{$uuid}->{$key};
    }
  }
  
  $globals{NAMESPACE} = $hash;
  return $hash;
}

sub host_namespace($)
{
  my $force = shift;
  
  ### Get host info
  return $globals{HOST_NAMESPACE} if ( !$force && $globals{HOST_NAMESPACE});
  
  my $hash;
  my ($fh,$buffer);
  my $command = $cmd_xe." host-list params 2>&1|";
  open($fh,$command) || die "Could not execute:\n".$command." ".$!;
  
  while ( <$fh> )
  {
    chomp;
    if ( my ($name,$value) = split(/\(.*\)\s*?:/,$_) )
    {
      $name =~ s/(^\s*|\s*$)//g;
      $value =~ s/(^\s*|\s*$)//g;
      $hash->{$name} = $value;
    }
  }
  close $fh;
  $globals{HOST_NAMESPACE} = $hash;
  return $globals{HOST_NAMESPACE};
}

sub vm_disk_list($)
{
  my $uuid = shift || die "Missing UUID";
  my ($fh,$buffer,$result,$hash,$key);
  my $command = $cmd_xe." vm-disk-list uuid=".$uuid." 2>&1|";
  open($fh,$command) || die "Could not execute:\n".$command." ".$!;
  
  while ( <$fh> )
  {
    chomp;
    if ( my ($name,$value) = split(/\(.*\)\s*?:/,$_) )
    {
      if ( $name =~ m/.*(Disk \d+).*VBD:/ ) { $key = $1;next; }
      if ( $name =~ m/.*(\d+).*VDI:/ ) { $key = ''; next; }
      $name =~ s/(^\s*|\s*$)//g;
      $value =~ s/(^\s*|\s*$)//g;
      if ( $key && $name && $value )
      {
        $hash->{$key}->{$name} = $value;
      }
    }
  }
  close $fh;
  return $hash;
  
}

sub get_domain_id_from_uuid($)
{
  my $uuid = shift || die "Missing UUID";
  my ($fh,$buffer);
  my $domain_id;
  
  my $command = $cmd_xe." vm-param-get param-name=dom-id uuid=".$uuid." 2>&1|";
  open($fh,$command) || die "Could not execute:\n".$command." ".$!;
  
  $domain_id = <$fh>;
  close $fh;
  
  chomp $domain_id;
  
  if ( $domain_id =~ m/^\d*$/ )
  {
    return $domain_id;
  } else {
    error("Couldn't get dom-id for uuid ".$uuid);
    return;
  }
}

sub xe_cmd($)
{
  my $cmd = shift || die "Missing UUID";
  my ($fh,$buffer,$result);
  
  my $command = $cmd_xe." ".$cmd." 2>&1|";
  open($fh,$command) || die "Could not execute:\n".$command." ".$!;
  
  $result = <$fh>;
  close $fh;
  
  return $result;
      
}

